@@ -1,8 +1,8 @@
 import calendar
-from datetime import datetime, timedelta as td, time
+from datetime import date, datetime, timedelta as td, time, timezone
 from enum import IntEnum
 
-import pytz
+UTC = timezone.utc
 
 RANGES = [
     frozenset(range(0, 60)),
@@ -10,25 +10,18 @@ RANGES = [
     frozenset(range(1, 32)),
     frozenset(range(1, 13)),
     frozenset(range(0, 8)),
-    frozenset(range(0, 60)),
 ]
 
 SYMBOLIC_DAYS = "SUN MON TUE WED THU FRI SAT".split()
 SYMBOLIC_MONTHS = "JAN FEB MAR APR MAY JUN JUL AUG SEP OCT NOV DEC".split()
 DAYS_IN_MONTH = [None, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
+FIELD_NAMES = ["minute", "hour", "day-of-month", "month", "day-of-week"]
 
 
 class CronSimError(Exception):
     pass
 
 
-def _int(value):
-    if not value.isdigit():
-        raise CronSimError("Bad value: %s" % value)
-
-    return int(value)
-
-
 class Field(IntEnum):
     MINUTE = 0
     HOUR = 1
@@ -36,16 +29,28 @@ class Field(IntEnum):
     MONTH = 3
     DOW = 4
 
+    def msg(self):
+        return "Bad %s" % FIELD_NAMES[self]
+
+    def _int(self, value):
+        if value == "":
+            raise CronSimError(self.msg())
+        for ch in value:
+            if ch not in "0123456789":
+                raise CronSimError(self.msg())
+
+        return int(value)
+
     def int(self, s):
-        if self == Field.MONTH and s.upper() in SYMBOLIC_MONTHS:
-            return SYMBOLIC_MONTHS.index(s.upper()) + 1
+        if self == Field.MONTH and s in SYMBOLIC_MONTHS:
+            return SYMBOLIC_MONTHS.index(s) + 1
 
-        if self == Field.DOW and s.upper() in SYMBOLIC_DAYS:
-            return SYMBOLIC_DAYS.index(s.upper())
+        if self == Field.DOW and s in SYMBOLIC_DAYS:
+            return SYMBOLIC_DAYS.index(s)
 
-        v = _int(s)
+        v = self._int(s)
         if v not in RANGES[self]:
-            raise CronSimError("Bad value: %s" % s)
+            raise CronSimError(self.msg())
 
         return v
 
@@ -59,20 +64,28 @@ class Field(IntEnum):
                 result.update(self.parse(term))
             return result
 
+        if self == Field.DOW and "L" in s:
+            value = s[:-1]
+            if not value.isdigit():
+                raise CronSimError(self.msg())
+
+            dow = self.int(s[:-1])
+            return {(dow, CronSim.LAST)}
+
         if "#" in s and self == Field.DOW:
             term, nth = s.split("#", maxsplit=1)
-            nth = _int(nth)
+            nth = self._int(nth)
             if nth < 1 or nth > 5:
-                raise CronSimError("Bad value: %s" % s)
+                raise CronSimError(self.msg())
 
             spec = (self.int(term), nth)
             return {spec}
 
         if "/" in s:
             term, step = s.split("/", maxsplit=1)
-            step = _int(step)
+            step = self._int(step)
             if step == 0:
-                raise CronSimError("Step cannot be zero")
+                raise CronSimError(self.msg())
 
             items = sorted(self.parse(term))
             if items == [CronSim.LAST]:
@@ -90,73 +103,68 @@ class Field(IntEnum):
             end = self.int(end)
 
             if end < start:
-                raise CronSimError("Range end cannot be smaller than start")
+                raise CronSimError(self.msg())
 
             return set(range(start, end + 1))
 
-        if self == Field.DAY and s in ("L", "l"):
+        if self == Field.DAY and s == "L":
             return {CronSim.LAST}
 
         return {self.int(s)}
 
 
-class NoTz(object):
-    def localize(self, dt, is_dst=None):
-        return dt
-
-    def normalize(self, dt):
-        return dt
+def is_imaginary(dt):
+    return dt != dt.astimezone(UTC).astimezone(dt.tzinfo)
 
 
 class CronSim(object):
     LAST = -1000
 
-    def __init__(self, expr, dt):
-        self.tz = dt.tzinfo or NoTz()
-        self.fixup_tz = None
+    def __init__(self, expr: str, dt: datetime):
         self.dt = dt.replace(second=0, microsecond=0)
 
-        parts = expr.split()
+        parts = expr.upper().split()
         if len(parts) != 5:
             raise CronSimError("Wrong number of fields")
 
+        # In Debian cron, if either the day-of-month or the day-of-week field
+        # starts with a star, then there is an "AND" relationship between them.
+        # Otherwise it's "OR".
+        self.day_and = parts[2].startswith("*") or parts[4].startswith("*")
+
         self.minutes = Field.MINUTE.parse(parts[0])
         self.hours = Field.HOUR.parse(parts[1])
         self.days = Field.DAY.parse(parts[2])
         self.months = Field.MONTH.parse(parts[3])
         self.weekdays = Field.DOW.parse(parts[4])
 
-        # If day is unrestricted but dow is restricted then match only with dow:
-        if self.days == RANGES[Field.DAY] and self.weekdays != RANGES[Field.DOW]:
-            self.days = set()
-
-        # If dow is unrestricted but day is restricted then match only with day:
-        if self.weekdays == RANGES[Field.DOW] and self.days != RANGES[Field.DAY]:
-            self.weekdays = set()
-
         if len(self.days) and min(self.days) > 29:
             # Check if we have any month with enough days
             if min(self.days) > max(DAYS_IN_MONTH[month] for month in self.months):
-                raise CronSimError("Bad day-of-month")
+                raise CronSimError(Field.DAY.msg())
 
-        if self.dt.tzinfo in (None, pytz.utc):
-            # No special DST handling for naive datetimes or UTC
+        self.fixup_tz = None
+        if self.dt.tzinfo in (None, UTC):
+            # No special DST handling for UTC
             pass
         else:
-            # Special handling for jobs that run at specific time, or with
-            # a granularity greater than one hour (to mimic Debian cron).
-            # Convert to naive datetime, will convert back to the tz-aware
-            # in __next__, right before returning the value.
             if not parts[0].startswith("*") and not parts[1].startswith("*"):
-                self.fixup_tz, self.tz = self.tz, NoTz()
+                # Will use special handling for jobs that run at specific time, or
+                # with a granularity greater than one hour (to mimic Debian cron).
+                self.fixup_tz = self.dt.tzinfo
                 self.dt = self.dt.replace(tzinfo=None)
 
-    def tick(self, minutes=1):
-        """ Roll self.dt forward by 1 or more minutes and fix timezone. """
+    def tick(self, minutes: int = 1) -> None:
+        """Roll self.dt forward by 1 or more minutes and fix timezone."""
 
-        self.dt = self.tz.normalize(self.dt + td(minutes=minutes))
+        if self.dt.tzinfo not in (None, UTC):
+            as_utc = self.dt.astimezone(UTC)
+            as_utc += td(minutes=minutes)
+            self.dt = as_utc.astimezone(self.dt.tzinfo)
+        else:
+            self.dt += td(minutes=minutes)
 
-    def advance_minute(self):
+    def advance_minute(self) -> bool:
         """Roll forward the minute component until it satisfies the constraints.
 
         Return False if the minute meets contraints without modification.
@@ -182,7 +190,7 @@ class CronSim(object):
 
         return True
 
-    def advance_hour(self):
+    def advance_hour(self) -> bool:
         """Roll forward the hour component until it satisfies the constraints.
 
         Return False if the hour meets contraints without modification.
@@ -202,26 +210,44 @@ class CronSim(object):
 
         return True
 
-    def match_day(self, d):
-        # Does the day of the month match?
+    def match_dom(self, d: date) -> bool:
+        """Return True is day-of-month matches."""
         if d.day in self.days:
             return True
 
-        if CronSim.LAST in self.days:
+        if self.LAST in self.days:
             _, last = calendar.monthrange(d.year, d.month)
             if d.day == last:
                 return True
 
-        # Does the day of the week match?
+        return False
+
+    def match_dow(self, d: date) -> bool:
+        """Return True is day-of-week matches."""
         dow = d.weekday() + 1
         if dow in self.weekdays or dow % 7 in self.weekdays:
             return True
 
+        if (dow, self.LAST) in self.weekdays or (dow % 7, self.LAST) in self.weekdays:
+            _, last = calendar.monthrange(d.year, d.month)
+            if d.day + 7 > last:
+                # Same day next week would be outside this month.
+                # So this is the last one this month.
+                return True
+
         idx = (d.day + 6) // 7
         if (dow, idx) in self.weekdays or (dow % 7, idx) in self.weekdays:
             return True
 
-    def advance_day(self):
+        return False
+
+    def match_day(self, d: date) -> bool:
+        if self.day_and:
+            return self.match_dom(d) and self.match_dow(d)
+
+        return self.match_dom(d) or self.match_dow(d)
+
+    def advance_day(self) -> bool:
         """Roll forward the day component until it satisfies the constraints.
 
         This method advances the date until it matches either the
@@ -243,11 +269,11 @@ class CronSim(object):
                 # This significantly speeds up the "0 0 * 2 MON#5" case
                 break
 
-        self.dt = self.tz.localize(datetime.combine(needle, time()))
+        self.dt = datetime.combine(needle, time(), tzinfo=self.dt.tzinfo)
         return True
 
-    def advance_month(self):
-        """Roll forward the month component until it satisfies the constraints. """
+    def advance_month(self) -> None:
+        """Roll forward the month component until it satisfies the constraints."""
 
         if self.dt.month in self.months:
             return
@@ -256,16 +282,22 @@ class CronSim(object):
         while needle.month not in self.months:
             needle = (needle.replace(day=1) + td(days=32)).replace(day=1)
 
-        self.dt = self.tz.localize(datetime.combine(needle, time()))
+        self.dt = datetime.combine(needle, time(), tzinfo=self.dt.tzinfo)
 
     def __iter__(self):
         return self
 
-    def __next__(self):
+    def __next__(self) -> datetime:
         self.tick()
 
+        start_year = self.dt.year
         while True:
             self.advance_month()
+            if self.dt.year > start_year + 50:
+                # Give up if there is no match for 50 years.
+                # It would be nice to detect "this will never yield any results"
+                # situations in a more intelligent way.
+                raise StopIteration
 
             if self.advance_day():
                 continue
@@ -277,15 +309,14 @@ class CronSim(object):
                 continue
 
             # If all constraints are satisfied then we have the result.
-            # The last step is to see if self.fixup_dst is set. If it is,
-            # localize self.dt and handle conflicts.
+            # The last step is to check if we need to fix up an imaginary
+            # or ambiguous date.
             if self.fixup_tz:
-                while True:
-                    try:
-                        return self.fixup_tz.localize(self.dt, is_dst=None)
-                    except pytz.AmbiguousTimeError:
-                        return self.fixup_tz.localize(self.dt, is_dst=True)
-                    except pytz.NonExistentTimeError:
-                        self.dt += td(minutes=1)
+                result = self.dt.replace(tzinfo=self.fixup_tz, fold=0)
+                while is_imaginary(result):
+                    self.dt += td(minutes=1)
+                    result = self.dt.replace(tzinfo=self.fixup_tz)
+
+                return result
 
             return self.dt
